/*
 * xzx    : Copyright (c) Des Herriott 1993, 1994
 * xcoupe : Copyright (c) Allan Skillman 1995.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the copyright holder not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Allan Skillman (Based on xzx coding by Des Herriott)
 *   */

/* Source file emul.c - this is the heart of the Z80 emulation. */

#include <stdio.h>
#include <sys/types.h>

#include "z80.h"
#include "mem.h"
#include "rtcfg.h"
#include "util.h"
#include "debug.h"
#include "xsam.h"

#ifndef __GNUC__
#define inline
#endif


int disass;

extern uns8 ixiy_tbl[];
extern uns8 parity_tbl[];
extern int hcarry_tbl[], sub_hcarry_tbl[];
extern int overflow_tbl[], sub_overflow_tbl[];

static uns16 work16;    	/* 16-bit temporary workspace */
static uns8  work8;     	/* 8-bit  temporary workspace */
static uns8  work8_2;		/* 8-bit  temporary secondary workspace */
static uns32 work32;    	/* 32-bit temporary workspace */
static int cin7;		/* carry-in on bit 7 */
static int idx;			/* for half-carry & overflow detection */

static uns8 high_nib, low_nib, add, carry, subflg;	/* used by DAA */

static sgn8 disp;		/* IX/IY displacement byte */
static int ixiy_op;
static uns16 *safe_hl;

int instrs;			/* no. of instructions executed since powerup */
int n_interrupts;		/* no. of interrupts since powerup */
#ifdef DEBUG
int last_instrs;
#endif
static int LineFreq;		/* No of instructions / line generated */
static int IntActiveTime;	/* Time (in instructions) an int is active (Port 249 <> 0xff) */

int LineNo; 			/* Scan line being generated */
int InterruptType;		/* Type of z80 interupt to deal with */

static int MouseActiveTime;	/* Time (in instructions) from a mouse strobe to a counter reset */

static int DelayedEI = 0;	/* Flag and counter to carry out a delayed */
static int EICount = 1;		/* EI as per Z80 specs */

/* Since the flags are stored in ints instead of in the f register itself,
 * we must store/retrieve them whenever the f register is read/written.
 * Fortunately, this doesn't happen much.
 */
inline void StoreFlags()
{
	if (carryFlag)    *f |= F_CARRY;    else *f &= ~F_CARRY;
	if (addsubFlag)   *f |= F_NADD;     else *f &= ~F_NADD;
	if (par_overFlag) *f |= F_OVERFLOW; else *f &= ~F_OVERFLOW;
	if (hcarryFlag)   *f |= F_HCARRY;   else *f &= ~F_HCARRY;
	if (zeroFlag)     *f |= F_ZERO;     else *f &= ~F_ZERO;
	if (signFlag)     *f |= F_NEG;      else *f &= ~F_NEG;
}

inline void RetrieveFlags()
{
	carryFlag    = *f & F_CARRY    ? 1 : 0;
	addsubFlag   = *f & F_NADD     ? 1 : 0;
	par_overFlag = *f & F_OVERFLOW ? 1 : 0;
	hcarryFlag   = *f & F_HCARRY   ? 1 : 0;
	zeroFlag     = *f & F_ZERO     ? 1 : 0;
	signFlag     = *f & F_NEG      ? 1 : 0;
}


/* Set the appropriate flags on an addition.  Return the new value
 * of the affected register.
 */
static inline void
add_and_test(uns8 reg2, uns8 carry)
{
	work16 = *a + reg2 + carry;
	idx = ((*a & 0x88) >> 1) | ((reg2 & 0x88) >> 2) | ((work16 & 0x88) >> 3);

	carryFlag = work16 & 0x0100;
	zeroFlag = !(work16 & 0xff);
	signFlag = work16 & 0x80;
	hcarryFlag = hcarry_tbl[idx & 0x07];
	par_overFlag = overflow_tbl[idx >> 4];
	Clr(addsubFlag);

	*a = (work16 & 0xff);
}

/* Set the appropriate flags on a subtraction.  Return the new value
 * of the affected register.
 */
static inline void
subtract_and_test(uns8 reg2, uns8 carry)
{
	work16 = *a - (reg2 + carry);
	idx = ((*a & 0x88) >> 1) | ((reg2 & 0x88) >> 2) | ((work16 & 0x88) >> 3);

	carryFlag = work16 & 0x0100;
	zeroFlag = !(work16 & 0xff);
	signFlag = work16 & 0x80;
	hcarryFlag = sub_hcarry_tbl[idx & 0x07];
	par_overFlag = sub_overflow_tbl[idx >> 4];
	Set(addsubFlag);

	*a = (work16 & 0xff);
}


/* Compare <val> with the accumulator.  Similar to subtraction,
 * but doesn't return anything, only sets some flags.
 */
static inline void
compare_and_test(uns8 val)
{
	work16 = *a - val;
	idx = ((*a & 0x88) >> 1) | ((val & 0x88) >> 2) | ((work16 & 0x88) >> 3);
	cin7 = ((*a & 0x7f) - (val & 0x7f)) & 0x80;

	carryFlag = work16 & 0x0100;
	zeroFlag = !(work16 & 0xff);
	signFlag = work16 & 0x80;
	hcarryFlag = sub_hcarry_tbl[idx & 0x07];
	par_overFlag = sub_overflow_tbl[idx >> 4];
	Set(addsubFlag);
}


/* The next two functions set the appropriate flags after an increment
 * or decrement of a given register <reg>.
 */
static inline void
inc_test(uns8 reg)
{
	par_overFlag = (reg == 0x80); 
	zeroFlag = !reg;
	signFlag = reg & 0x80;
	hcarryFlag = ((reg & 0x0F) == 0);
	Clr(addsubFlag);
}


static inline void
dec_test(uns8 reg)
{
	par_overFlag = (reg == 0x7f);
	zeroFlag = !reg;
	signFlag = reg & 0x80;
	hcarryFlag = ((reg & 0x0F) == 0x0F);
	Set(addsubFlag);
}


/* Perform a 16-bit addition, setting the appropriate flags.
 * <carry> must be either 1 or 0 on entry.
 */
static inline void
add_and_test16(uns16 reg2, int carry)
{
	work32 = *hl + reg2 + carry;
	idx = ((*hl & 0x8800) >> 9) | ((reg2 & 0x8800) >> 10) |
		  ((work32 & 0x8800) >> 11);

	carryFlag = work32 & 0x10000;
	zeroFlag = !(work32 & 0xffff);
	hcarryFlag = hcarry_tbl[idx & 0x07];
	par_overFlag = overflow_tbl[idx >> 4];
	signFlag = work32 & 0x8000;
	Clr(addsubFlag);

	*hl = (work32 & 0xffff);
}

/* Perform a 16-bit subtraction, setting the appropriate flags.
 * <carry> must be either 1 or 0 on entry.
 */
static inline void
subtract_and_test16(uns16 reg2, int carry)
{
	work32 = *hl - (reg2 + (uns16) carry);
	idx = ((*hl & 0x8800) >> 9) | ((reg2 & 0x8800) >> 10) |
		  ((work32 & 0x8800) >> 11);

	carryFlag = work32 & 0x10000;
	zeroFlag = !(work32 & 0xffff);
	hcarryFlag = sub_hcarry_tbl[idx & 0x07];
	par_overFlag = sub_overflow_tbl[idx >> 4];
	signFlag = work32 & 0x8000;
	Set(addsubFlag);

	*hl = (work32 & 0xffff);
}


/* Set the flags after the 3 bitwise operations.  These three
 * functions are very similar (or & xor are identical), but I
 * kept them separate for the sake of clarity.
 */
static inline void
do_and_flags(uns8 reg)
{
	zeroFlag = !reg; signFlag = reg & 0x80;
	par_overFlag = parity_tbl[reg];
	Set(hcarryFlag); Clr(carryFlag); Clr(addsubFlag);
}

static inline void
do_or_flags(uns8 reg)
{
	zeroFlag = !reg; signFlag = reg & 0x80;
	par_overFlag = parity_tbl[reg];
	Clr(hcarryFlag); Clr(carryFlag); Clr(addsubFlag);
}

static inline void
do_xor_flags(uns8 reg)
{
	zeroFlag = !reg; signFlag = reg & 0x80;
	par_overFlag = parity_tbl[reg];
	Clr(hcarryFlag); Clr(carryFlag); Clr(addsubFlag);
}


/* Set flags after a shift or rotate operation */
static inline void
do_shift_test(uns8 reg)
{
	Clr(hcarryFlag); Clr(addsubFlag);
	par_overFlag = parity_tbl[reg];
	zeroFlag = !reg; signFlag = reg & 0x80;
}

/* Set flags after an in r,(c) instruction [NB not in a,(n)] */
static inline void
do_input_flags(uns8 reg)
{
	Clr(hcarryFlag); Clr(addsubFlag);
	par_overFlag = parity_tbl[reg];
	zeroFlag = !reg; signFlag = reg & 0x80;
}



#ifdef SLOWDOWN
/* Do nothing! */
static void
nop(void)
{
}
#endif


static void
init_cpu(void)
{
	af = &(theProcessor.af_pair);
	hl = &(theProcessor.hl_pair);
	bc = &(theProcessor.bc_pair);
	de = &(theProcessor.de_pair);
	ix = &(theProcessor.ix_reg);
	iy = &(theProcessor.iy_reg);

#ifdef LITTLE_ENDIAN
	a  = (uns8 *)&(theProcessor.af_pair) + 1;
	f  = (uns8 *)&(theProcessor.af_pair);
	h  = (uns8 *)&(theProcessor.hl_pair) + 1;
	l  = (uns8 *)&(theProcessor.hl_pair);
	b  = (uns8 *)&(theProcessor.bc_pair) + 1;
	c  = (uns8 *)&(theProcessor.bc_pair);
	d  = (uns8 *)&(theProcessor.de_pair) + 1;
	e  = (uns8 *)&(theProcessor.de_pair);
#else
	a  = (uns8 *)&(theProcessor.af_pair);
	f  = (uns8 *)&(theProcessor.af_pair) + 1;
	h  = (uns8 *)&(theProcessor.hl_pair);
	l  = (uns8 *)&(theProcessor.hl_pair) + 1;
	b  = (uns8 *)&(theProcessor.bc_pair);
	c  = (uns8 *)&(theProcessor.bc_pair) + 1;
	d  = (uns8 *)&(theProcessor.de_pair);
	e  = (uns8 *)&(theProcessor.de_pair) + 1;
#endif

	carryFlag =  addsubFlag = par_overFlag = 0;
	hcarryFlag = zeroFlag =   signFlag = 0;
}


static void
reset_cpu(void)
{
	theProcessor.im = 0;
	theProcessor.iff1 = 0;
	theProcessor.iff2 = 0;

	init_pages();

}

/*
 * Do The Biz.
 */
void
xcoupe_main_loop(void)
{
	uns8 op;
	int refresh_interval;



#ifdef SLOWDOWN
	int i;
#endif
	register int instrs;			/* no. of instructions executed since powerup */
	register int IntActiveCounter;		/* Counter used for int Active timing */
	register int MouseActiveCounter;	/* Counter used for mouse Active timing */

	/* The rest of the registers are defined here */
	
	register uns16 pc, sp;
	register uns8  refresh, int_vec;

	/* from xsam.c */
	extern void (*screen_refresh)(void);
	extern int clut_change_in_frame;
	extern int border_change_in_frame;
	extern int vmpr_change_in_frame;
	extern int DoPaletteSwitching;

	extern void update_line(void);
	extern void update_to_first_line(void);
	extern void update_border_to_first_line(void);
	extern void check_events(void);
	
	/* from mouse.c */
	
	extern int MouseStrobed;
	

	init_cpu();
	reset_cpu();
	sp = pc = 0;
	int_vec  = refresh  = 0;
		
	InterruptType = 0;

	disass = 0;

#ifdef MITSHM
	if (GetCfg(use_mitshm))
		refresh_interval = GetCfg(rr_shm);
	else
#endif
		refresh_interval = GetCfg(rr_noshm);

	/* start interrupts */

	/* Calculate the mean no. of instructions/line generated */

	LineFreq = GetCfg(ifreq) / (60. + 192.);

	/* Calculate the mean no. of instructions executed in the 100usecs
	   an interrupt remains active */
	   
	IntActiveTime = GetCfg(ifreq) * (100.e-6 / 0.02);

	xcoupe_mesg(XCOUPE_INFO,"LineFreq :%d\nIntActiveTime : %d\n",
		LineFreq, IntActiveTime);
	LineNo   = 0;	
	IntActiveCounter = IntActiveTime;

	/* Similar for mouse reset */	
	MouseActiveTime = IntActiveTime;
	MouseActiveCounter = MouseActiveTime;

	/* The main Z80 emulation loop */
	while(1) {

		instrs++;

		if (disass) {
			disass = run_debug(pc);
			if (disass == 2){
				fprintf(stderr, "Z80 CPU State\n-------------\n");

				StoreFlags();

				fprintf(stderr, "  AF  = %04X\t\t  AF' = %04X\t\t  A = %02X  F = %02X\n",
				theProcessor.af_pair, theProcessor.af_alt, *a, *f);
				fprintf(stderr, "  BC  = %04X\t\t  BC' = %04X\t\t  B = %02X  C = %02X\n",
				theProcessor.bc_pair, theProcessor.bc_alt, *b, *c);
				fprintf(stderr, "  DE  = %04X\t\t  DE' = %04X\t\t  D = %02X  E = %02X\n",
				theProcessor.de_pair, theProcessor.de_alt, *d, *e);
				fprintf(stderr, "  HL  = %04X\t\t  HL' = %04X\t\t  H = %02X  L = %02X\n",
				theProcessor.hl_pair, theProcessor.hl_alt, *h, *l);
				fprintf(stderr, "  IX  = %04X\t\t  IY  = %04X\n\n",
				theProcessor.ix_reg, theProcessor.iy_reg);

				fprintf(stderr, "  PC  = %04X\t\t  SP  = %04X\n",
				pc, sp);

				fprintf(stderr, "   R  = %02X\t\t   I  = %02X\n",
				refresh, int_vec);

				fprintf(stderr, "  interrupts %s, IM%d\n",
				theProcessor.iff1 ? "enabled" : "disabled", theProcessor.im);

				fprintf(stderr, "  FLAGS: C=%d N=%d P/V=%d H=%d Z=%d S=%d\n\n",
				!!carryFlag, !!addsubFlag, !!par_overFlag, 
				!!hcarryFlag, !!zeroFlag, !!signFlag);
				}

		}

#ifdef SLOWDOWN
		for (i = 0; i < GetCfg(slowdown); i++) nop();
#endif

		op = GetNextOp;
restart_switch:

		switch (op) {
#			include "z80ops.c"
		}
		if (ixiy_op) {
			ixiy_op = 0; *hl -= disp; hl = safe_hl;
#ifdef LITTLE_ENDIAN
			h = (uns8 *)hl + 1; l = (uns8 *)hl;
#else
			h = (uns8 *)hl; l = (uns8 *)hl + 1;
#endif
		}

		/* Sort out the EI instruction - enable the interrupts 
		   AFTER  the next instruction */
		if (DelayedEI && !EICount--) {
			theProcessor.iff1 = 1;
			theProcessor.iff2 = 1;
			DelayedEI = 0;
			EICount = 1;			
		}

		/* If we have strobed the mouse, we will need to do
		   a reset after 100 usecs */
		   
		if ((MouseStrobed) && (!MouseActiveCounter--)) {
			MouseActiveCounter = MouseActiveTime;
			MouseStrobed = 0;
			mouse_reset(0);
		}


		/* On average there are about 10000 Z80 instructions in 0.02 secs, on
	  	 * a 6.0MHz Z80b (assuming average instruction length of 12 T-states)
		 */

		/* Call the interrupt if required */

		if ((status_reg != 0xff) && theProcessor.iff1) {			
			/*do_interrupt(uns8 op) */
			/* kludge: make sure we don't return to a halt instruction */
			if (op == 0x76) (pc)++;

			/* Disable the interrupts */
			theProcessor.iff1 = 0;
			theProcessor.iff2 = 0;
			PushValue(pc);

			if (theProcessor.im == 2) {
				pc = read_word((int_vec << 8) | 0xff);
			} else {
				pc = 0x38;
			}
		}

		if ((status_reg != 0xff) && (!IntActiveCounter--)) {
			IntActiveCounter = IntActiveTime;
			status_reg = 0xff;
		}
		

		if (!(instrs % LineFreq)) {
			
			if (LineNo == (191+60)) {
			/* FlyBack - Frame interrupt */

				LineNo = 0;
				IntActiveCounter = IntActiveTime;
				status_reg = 0xf7;
				
				n_interrupts++;
				
				refresh = instrs & 0x7f;

				check_events();

				/* every 25 interrupts -> twice/sec. */
				if (((n_interrupts % 25) == 0) && (VMPR_SCREEN_MODE < MODE_3))
					do_flashing();

				if (DoPaletteSwitching) {
					update_to_first_line(); 
					clut_change_in_frame = 0;
				}
				if (!GetCfg(NoBorder) && border_change_in_frame) {
					update_border_to_first_line();
					border_change_in_frame = 0;
				}
				if ((n_interrupts % refresh_interval) == 0)
					screen_refresh();
				
			}
			else if ((line_int < 192) && (LineNo == line_int+30)) {
			/* Line Interrupt */

				IntActiveCounter = IntActiveTime;
				status_reg = 0xfe;
			}
			

			/* Now we are Generating the next line down so
			   update the counter */
			   
			LineNo++;
			   
			if (LineNo <= 29) {
				hpen = 0;
				attr_reg = 255;
			}
			else if (LineNo > 29 && LineNo < 222) {
			/* If we had a clut change in the frame do the update */
				if (DoPaletteSwitching && clut_change_in_frame) 
					update_line();

				/* Update the lpen port */
				hpen = LineNo - 30;			
				attr_reg = 0;
			}
			else {
				hpen = 192;
				/* Set attr port to 0xff in the border for
				   Spectrum compatability */
				attr_reg = 255;
			}
		
	        	switch(InterruptType) {
		    	   case Z80_none:
		    		break;

	            	   case Z80_nmi:
				if (read_byte(pc) == 0x76)
					(pc)++;
				PushValue(pc);
				theProcessor.iff2 = theProcessor.iff1;
				theProcessor.iff1 = 0;
				pc = 0x0066;
				InterruptType = Z80_none;
				break;
	            	   case Z80_reset:
				init_pages();
	       	                 init_ports();
				init_cpu();
				reset_cpu();
				InterruptType = 0;
				break;
		    	   default:
		    		break;
			}
		}

	} 
}
				
